import json
import pathlib

from hidori_core.modules import MODULES_REGISTRY
from hidori_core.utils import Messenger

RETURN_TO_INITIATOR_MODULES = ["file-push"]


def main():
    tasks_path = pathlib.Path(__file__).parent / "tasks.json"
    with open(tasks_path) as tasks_file:
        data = json.load(tasks_file)

    # TODO: Add pre-flight env detection and verification
    # python >= 3.9
    # target python >= 3.7
    # systemd
    # broaden access and lower reqs

    savefile_path = pathlib.Path(__file__).parent / ".savefile"
    starting_step = 0
    if savefile_path.exists():
        with savefile_path.open() as f:
            starting_step = int(f.read())

    for (step, (task, task_data)) in enumerate(data.items()):
        if step < starting_step:
            continue

        if step > starting_step:
            if task_data["module"] in RETURN_TO_INITIATOR_MODULES:
                with savefile_path.open("w") as f:
                    f.write(str(step))
                return

        module = MODULES_REGISTRY[task_data["module"]]
        messenger = Messenger(task)
        module.validate_and_execute(task_data, messenger)
        messenger.say_all()

    # TODO: Summary of failed, affected and unaffected tasks


if __name__ == "__main__":
    main()



from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from hidori_core.schema.errors import (
    ConstraintError,
    SchemaError,
    SkipFieldError,
    ValidationError,
)

TField = TypeVar("TField", bound="Field")

FIELDS_REGISTRY: List[Type["Field"]] = []


class Field:
    required: bool

    def __init_subclass__(cls) -> None:
        FIELDS_REGISTRY.append(cls)

    @classmethod
    def from_annotation(
        cls: Type[TField], annotation: Any, required: bool = True
    ) -> Optional[TField]:
        ...

    def validate(self, value: Optional[Any]) -> Optional[Any]:
        if self.required and not value:
            raise ValidationError("value for required field not provided")
        elif self.required is False and not value:
            raise SkipFieldError()
        else:
            return value


class Constraint:
    def process_schema(self, annotations: Dict[str, Any]) -> None:
        ...

    def apply(self, schema: Type["Schema"], data: Dict[str, Any]) -> None:
        ...


class Schema:
    fields: Dict[str, Field]

    def __init_subclass__(cls):
        cls.fields = {}
        errors = {}

        for name, annotation in cls.__annotations__.items():
            if name == "fields":
                continue

            constraint = getattr(cls, name, None)
            if constraint and isinstance(constraint, Constraint):
                try:
                    constraint.process_schema(cls.__annotations__)
                except ConstraintError as e:
                    errors[name] = str(e)
                    continue

            cls.fields[name] = field_from_annotation(annotation)

        if errors:
            raise SchemaError(errors)

    @classmethod
    def validate(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        validated_data: Dict[str, Any] = {}
        errors = {}

        for name, field in cls.fields.items():
            constraint = getattr(cls, name, None)
            if constraint and isinstance(constraint, Constraint):
                constraint.apply(cls, data)

            try:
                validated_data[name] = field.validate(data.get(name))
            except ValidationError as e:
                errors[name] = str(e)
                continue
            except SkipFieldError:
                continue

        if errors:
            raise SchemaError(errors)

        return validated_data


def field_from_annotation(annotation: Any, required: bool = True) -> Field:
    for field_cls in FIELDS_REGISTRY:
        field = field_cls.from_annotation(annotation, required)
        if field is not None:
            return field

    if annotation.__origin__ == Union:
        # TODO: Impl unions other than Optional when necessary
        assert len(annotation.__args__) == 2

        required = False
        for base_type in annotation.__args__:
            if isinstance(None, base_type):
                continue

            return field_from_annotation(base_type, required)

    # TODO: Finish impl for other cases
    raise RuntimeError("internal error")



from typing import Any, Optional, Tuple

from hidori_core.compat.typing import Literal
from hidori_core.schema.base import Field, ValidationError


class Text(Field):
    @classmethod
    def from_annotation(
        cls, annotation: Any, required: bool = True
    ) -> Optional["Text"]:
        return cls(required) if annotation == str else None

    def __init__(self, required: bool) -> None:
        self.required = required


class OneOf(Field):
    @classmethod
    def from_annotation(
        cls, annotation: Any, required: bool = True
    ) -> Optional["OneOf"]:
        if annotation.__origin__ == Literal:
            return cls(annotation.__args__, required)

        return None

    def __init__(self, values: Tuple[Any], required: bool) -> None:
        self.allowed_values = values
        self.required = required

    def validate(self, value: Optional[Any]) -> Optional[Any]:
        super().validate(value)

        if value in self.allowed_values:
            return value
        else:
            allowed_values = ", ".join(self.allowed_values)
            raise ValidationError(
                f"value `{value}` not allowed; allowed values are `{allowed_values}`"
            )



import os
from typing import Dict

from hidori_core.modules.base import Module
from hidori_core.schema import Schema
from hidori_core.utils import Messenger


class HelloSchema(Schema):
    ...


class HelloModule(Module, name="hello", schema_cls=HelloSchema):
    def execute(self, validated_data: Dict[str, object], messenger: Messenger):
        uname_result = os.uname()
        messenger.queue_success(
            f"Hello from {uname_result.sysname} {uname_result.nodename} "
            f"{uname_result.release}"
        )
        return {"state": "unaffected"}



from hidori_cli.apps.base import BaseCLIApplication


class HidoriPipelineApplication(BaseCLIApplication):
    """hidori-pipeline CLI."""


def main() -> None:
    app = HidoriPipelineApplication()
    app.run()



import json
import pathlib
import shutil
import subprocess
import tempfile
from typing import Any

import tomllib

import hidori_core
import hidori_runner

SSH_OPTIONS = [
    "-o ControlMaster=auto",
    "-o ControlPath=~/.ssh/control-%r@%h:%p",
    "-o ControlPersist=yes",
]

PIPELINE_MODULES_REGISTRY: dict[str, type["PipelineStep"]] = {}

# TODO: Pipelines don't belong in hidori_core


class PipelineStep:
    def __init_subclass__(cls, /, module_name: str) -> None:
        super().__init_subclass__()

        if module_name in PIPELINE_MODULES_REGISTRY:
            raise RuntimeError(f"{module_name} module name is already registered.")
        PIPELINE_MODULES_REGISTRY[module_name] = cls

    def __init__(self, pipeline: "Pipeline") -> None:
        self._pipeline = pipeline

    def run(self) -> None:
        raise NotImplementedError()


class DefaultPipelineStep(PipelineStep, module_name="*"):
    def run(self):
        if self._pipeline.running:
            return

        self._pipeline.running = True
        self._pipeline.run_remote_modules()


class FilePushPipelineStep(PipelineStep, module_name="file-push"):
    def run(self):
        self._pipeline.running = True
        self._pipeline.run_remote_modules()


class Pipeline:
    @classmethod
    def from_toml_path(cls, path: str) -> "Pipeline":
        with open(path, "rb") as f:
            data = tomllib.load(f)

        return Pipeline(data)

    def __init__(self, data: dict[str, Any]) -> None:
        self.prepared = False
        self.running = False
        self._executor_dir = None
        self.steps: list[PipelineStep] = []

        self._data = data
        # TODO: Actually validate the provided data
        # TODO: Replace hardcoded vm host data
        # TODO: Add support for driver config val
        # TODO: Add support for remote_path config val
        # TODO: Extract driver-specific code to the drivers
        host_data = data["hosts"]["vm"]
        self.ssh_ip = host_data["ip"]
        self.ssh_user = host_data["user"]

        modules = [task["module"] for _, task in data["tasks"].items()]
        for module_name in modules:
            self.steps.append(
                PIPELINE_MODULES_REGISTRY.get(
                    module_name, PIPELINE_MODULES_REGISTRY["*"]
                )(self)
            )

    def prepare(self):
        # TODO: Remote runner should only take required modules.
        # Use MODULES_REGISTRY and self.modules for that
        # It will also allow third parties to define their own modules.
        # TODO: This should likely be part of the SSH driver
        with tempfile.TemporaryDirectory(prefix="hidori-") as tmpdirpath:
            # TODO: This is terrible. Delete as soon as 3.7 drops.
            import typing_extensions

            ty_exts_path = typing_extensions.__file__
            tmp_ty_exts_path = pathlib.Path(tmpdirpath) / "typing_extensions.py"
            shutil.copyfile(ty_exts_path, tmp_ty_exts_path)

            self._copy_core_tree("compat", tmpdirpath)
            self._copy_core_tree("modules", tmpdirpath)
            self._copy_core_tree("schema", tmpdirpath)
            self._copy_core_tree("utils", tmpdirpath)

            # TODO: Use appropriate executor instead of a hardcoded remote
            executor_path = (
                pathlib.Path(hidori_runner.__file__).parent / "executors/remote.py"
            )
            tmp_executor_path = pathlib.Path(tmpdirpath) / "executor.py"
            shutil.copyfile(executor_path, tmp_executor_path)

            tmp_tasks_path = pathlib.Path(tmpdirpath) / "tasks.json"
            with open(tmp_tasks_path, "w") as tasks_file:
                json.dump(self._data["tasks"], tasks_file)

            # TO THE STARS!
            # TODO: Part of the transport
            scp_cmd = (
                f"scp {' '.join(SSH_OPTIONS)} -qr {tmpdirpath} "
                f"{self.ssh_user}@{self.ssh_ip}:{tmpdirpath}"
            )
            subprocess.run(scp_cmd.split())

        self._executor_dir = tmpdirpath
        self._prepared = True

    def _copy_core_tree(self, dirname: str, dest: str) -> None:
        tmp_core_init_path = pathlib.Path(hidori_core.__file__).parent / "__init__.py"
        open(tmp_core_init_path, "w").close()
        core_package_path = pathlib.Path(hidori_core.__file__).parent / dirname
        tmp_core_package_path = pathlib.Path(dest) / f"hidori_core/{dirname}"
        shutil.copytree(
            str(core_package_path), str(tmp_core_package_path), dirs_exist_ok=True
        )

    def run(self):
        if not self._prepared:
            raise RuntimeError("pipeline is not prepared")

        for pipeline_step in self.steps:
            pipeline_step.run()

    def run_remote_modules(self):
        # TODO: Rename this method - modules won't always run on remote.
        # TODO: Impl a dedicated ssh Transport class
        # TODO: Replace subprocess calls
        runner_ssh_cmd = (
            f"ssh {' '.join(SSH_OPTIONS)} -qt {self.ssh_user}@{self.ssh_ip} "
            f"python3 {self._executor_dir}/executor.py"
        )
        subprocess.run(runner_ssh_cmd.split())



from typing import Any, Dict, Type

from hidori_core.schema.base import Schema
from hidori_core.schema.errors import SchemaError
from hidori_core.utils.messenger import Messenger

MODULES_REGISTRY: Dict[str, "Module"] = {}


class Module:
    def __init_subclass__(cls, *, name: str, schema_cls: Type[Schema]) -> None:
        super().__init_subclass__()

        if name in MODULES_REGISTRY:
            raise RuntimeError(f"{name} module is already registered.")
        MODULES_REGISTRY[name] = cls(schema_cls)

    def __init__(self, schema_cls: Type[Schema]) -> None:
        self.schema = schema_cls()

    def validate_and_execute(
        self, task_data: Dict[str, Any], messenger: Messenger
    ) -> Dict[str, str]:
        try:
            validated_data = self.schema.validate(task_data)
        except SchemaError as e:
            module = task_data["module"]
            for field, error in e.errors.items():
                messenger.queue_error(f"module::{module}, field::{field}: {error}")
            return {"state": "error"}

        return self.execute(validated_data, messenger)

    def execute(
        self, validated_data: Dict[str, Any], messenger: Messenger
    ) -> Dict[str, str]:
        raise NotImplementedError()



from dataclasses import dataclass, field

from hidori_cli.commands.base import BaseData, Command
from hidori_core.pipeline import Pipeline


@dataclass
class PipelineRunData(BaseData):
    # TODO: Filepath type with validation that file exists?
    pipeline_file: str = field(metadata={"help": "Path to the TOML pipeline file"})


class PipelineRunCommand(Command[PipelineRunData]):
    """pipeline-run command"""

    data_cls = PipelineRunData

    def execute(self, data: PipelineRunData) -> None:
        pipeline = Pipeline.from_toml_path(data.pipeline_file)
        pipeline.prepare()
        pipeline.run()



import datetime


class Colors:
    RED = "\033[31m"
    GREEN = "\033[32m"
    BLUE = "\033[34m"
    RESET = "\033[39m"


class Modifiers:
    BOLD = "\033[1m"
    RESET = "\033[0m"


class Messenger:
    def __init__(self, task: str) -> None:
        # TODO: user and host are hardcoded
        self._user = "root"
        self._host = "vm"
        self._task = task
        self._messages: list[str] = []

    def queue_success(self, message: str) -> None:
        now = datetime.datetime.now().time().strftime("%H:%M:%S")
        self._messages.append(
            f"[{now}] {Modifiers.BOLD}{Colors.GREEN}"
            f"OK:{Colors.RESET}{Modifiers.RESET} {message}"
        )

    def queue_error(self, message: str) -> None:
        now = datetime.datetime.now().time().strftime("%H:%M:%S")
        self._messages.append(
            f"[{now}] {Modifiers.BOLD}{Colors.RED}"
            f"ERROR:{Colors.RESET}{Modifiers.RESET} {message}"
        )

    def queue_affected(self, message: str) -> None:
        now = datetime.datetime.now().time().strftime("%H:%M:%S")
        self._messages.append(
            f"[{now}] {Modifiers.BOLD}{Colors.BLUE}"
            f"AFFECTED:{Colors.RESET}{Modifiers.RESET} {message}"
        )

    def queue_info(self, message: str) -> None:
        now = datetime.datetime.now().time().strftime("%H:%M:%S")
        self._messages.append(
            f"[{now}] {Modifiers.BOLD}" f"INFO:{Modifiers.RESET} {message}"
        )

    def say_all(self):
        print(
            f"{Modifiers.BOLD}[{self._user}@{self._host}: "
            f"{self._task}]{Modifiers.RESET}"
        )
        for message in self._messages:
            print(message)
        print()



from typing import Dict


class ValidationError(Exception):
    ...


class ConstraintError(Exception):
    ...


class SkipFieldError(Exception):
    ...


class SchemaError(Exception):
    def __init__(self, errors: Dict[str, str]) -> None:
        self.errors = errors
        super().__init__(str(errors))



from hidori_cli.commands.base import COMMAND_REGISTRY, Command
from hidori_cli.commands.pipeline_run import PipelineRunCommand

__all__ = [
    "COMMAND_REGISTRY",
    "Command",
    "PipelineRunCommand",
]



from typing import Dict

import dbus

from hidori_core.modules.base import Module
from hidori_core.schema import Schema
from hidori_core.utils import Messenger


class HostnameSchema(Schema):
    name: str


class HostnameModule(Module, name="hostname", schema_cls=HostnameSchema):
    def execute(
        self, validated_data: Dict[str, str], messenger: Messenger
    ) -> Dict[str, str]:
        new_hostname = validated_data["name"]

        sysbus = dbus.SystemBus()
        hostnamed_proxy = sysbus.get_object(
            "org.freedesktop.hostname1", "/org/freedesktop/hostname1"
        )

        dbus_iface = dbus.Interface(hostnamed_proxy, "org.freedesktop.DBus.Properties")
        old_hostname = dbus_iface.Get("org.freedesktop.hostname1", "StaticHostname")
        if old_hostname == new_hostname:
            messenger.queue_success(f"hostname already set to {new_hostname}")
            return {"state": "unaffected"}

        hostname_iface = dbus.Interface(hostnamed_proxy, "org.freedesktop.hostname1")
        hostname_iface.SetStaticHostname(new_hostname, False)

        messenger.queue_affected(f"hostname changed to {new_hostname}")
        return {"state": "affected"}



from hidori_core.modules.base import MODULES_REGISTRY
from hidori_core.modules.dnf import DnfModule
from hidori_core.modules.hello import HelloModule
from hidori_core.modules.hostname import HostnameModule

__all__ = ["MODULES_REGISTRY", "DnfModule", "HelloModule", "HostnameModule"]



import argparse
import re

from hidori_cli.commands import COMMAND_REGISTRY, Command
from hidori_cli.commands.base import BaseData


class BaseCLIApplication:
    def __init__(self) -> None:
        name_parts: list[str] = re.findall(".[^A-Z]*", type(self).__name__)[:-1]
        self.name = "-".join([p.lower() for p in name_parts])
        self.parser = argparse.ArgumentParser(description=self.__doc__)
        self._commands: dict[str, Command[BaseData]] = {}
        self._register_commands()

    def _register_commands(self) -> None:
        subparsers = self.parser.add_subparsers(dest="subparser_name")
        for command_cls in COMMAND_REGISTRY[self.name].values():
            subparser = subparsers.add_parser(
                command_cls.name, help=command_cls.__doc__
            )
            self._commands[command_cls.name] = command_cls(subparser)

    def run(self) -> None:
        parser_data = self.parser.parse_args()
        subparser_name: str = parser_data.subparser_name
        command = self._commands[subparser_name]

        data = command.data_cls(**parser_data.__dict__)
        command.execute(data)



from typing import Any, Callable, Dict, List, Optional, Type

from hidori_core.schema.base import Constraint, Schema
from hidori_core.schema.errors import ConstraintError

DataCondtion = Callable[[Dict[str, Any]], bool]


class RequiresConstraint(Constraint):
    def __init__(
        self,
        field_names: List[str],
        data_conditions: Optional[List[DataCondtion]] = None,
    ) -> None:
        self.required_field_names = set(field_names)
        self.data_conditions = data_conditions or []

    def process_schema(self, annotations: Dict[str, Any]) -> None:
        undefined_field_names = ", ".join(
            self.required_field_names.difference(annotations.keys())
        )
        if undefined_field_names:
            raise ConstraintError(
                f"fields named `{undefined_field_names}` "
                "might be required but are undefined"
            )

    def apply(self, schema: Type[Schema], data: Dict[str, Any]) -> None:
        # TODO: data_conditions should likely move to the Constraint base cls
        for condition in self.data_conditions:
            if not condition(data):
                return

        for field_name in self.required_field_names:
            schema.fields[field_name].required = True


def Requires(
    field_names: List[str], data_conditions: Optional[List[DataCondtion]] = None
) -> Any:
    return RequiresConstraint(field_names, data_conditions)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'hidori_cli': 'src/hidori_cli',
 'hidori_cli.apps': 'src/hidori_cli/apps',
 'hidori_cli.commands': 'src/hidori_cli/commands',
 'hidori_runner': 'src/hidori_runner',
 'hidori_runner.drivers': 'src/hidori_runner/drivers',
 'hidori_runner.executors': 'src/hidori_runner/executors',
 'hidori_runner.transports': 'src/hidori_runner/transports'}

packages = \
['hidori_cli',
 'hidori_cli.apps',
 'hidori_cli.commands',
 'hidori_core',
 'hidori_core.compat',
 'hidori_core.modules',
 'hidori_core.schema',
 'hidori_core.utils',
 'hidori_runner',
 'hidori_runner.drivers',
 'hidori_runner.executors',
 'hidori_runner.transports']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.3.0,<5.0.0']

entry_points = \
{'console_scripts': ['hidori-pipeline = hidori_cli.apps.pipeline:main']}

setup_kwargs = {
    'name': 'hidori',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Hidori\n\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/hidori-dev/hidori/main.svg)](https://results.pre-commit.ci/latest/github/hidori-dev/hidori/main)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/hidori-dev/hidori)\n\nHidori is a modern, agentless, zero-dependency[^1][^2] variation on updating system state. General rule of thumb is that changes to the system are only done if the requested state is different from the actual state. Hidori modules are idempotent if the system is already in the desired state.\nEvery change in the target system is reported through the modules log as "affected".\n\nHidori communicates with a target machine through appropriate protocol that is designated by the chosen driver.\n\n## Install\n\nDepending on your environment you might wish to install Hidori globally or locally using chosen Python dependency manager such as poetry or pip.\nUltimately, the choice is yours. Either way, Hidori is safe for your Python environment - it does not pull any dependencies from PyPI[^2].\n\nExample using pip:\n```sh\npip install hidori\n```\n\n## Hello World\n\nWith Hidori installed you need a single TOML file that provides where and what should be done. A simple \'hello world\' example assuming that some machine is available at the given IP address:\n\n```toml\n[hosts]\n\n  [hosts.vm]\n  ip = "192.168.122.31"\n  user = "root"\n\n[tasks]\n\n  [tasks."Say hello"]\n  module = "hello"\n```\n\nNow you just need to run it (example assumes the file is named `pipeline.toml`):\n\n```sh\nhidori-pipeline run pipeline.toml\n```\n\nThe result on my machine is:\n\n```\n[root@vm: Say hello]\n[16:03:19] OK: Hello from Linux debian 4.19.0-21-amd64\n```\n\n## Support\n\nIn general, Hidori is based on Python 3.11, but `hidori_core` runs with any version of Python that is still supported.\nReason for that is vast majority of target systems don\'t have the latest Python runtime installed, so therefore `hidori_core`\nwhich includes all the code that runs on a target system can be expected to be supported according to the following table:\n\n| Python Version |     EOL Date     |\n| -------------- | ---------------- |\n| 3.7            | July 2023        |\n| 3.8            | November 2024    |\n| 3.9            | November 2025    |\n| 3.10           | November 2026    |\n| 3.11           | November 2027(?) |\n\n## Development\n\nThe dev environment can be setup with poetry:\n```sh\npoetry install\n```\n\n[^1]: Except for the necessary runtime - python, and system libraries that are used by modules\n[^2]: typing-extensions are necessary until at least July 2023 when Python 3.7\'s support will end',
    'author': 'Piotr Szpetkowski',
    'author_email': 'piotr.szpetkowski@pyquest.space',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, Optional

try:
    import dnf
except ImportError:
    dnf = None

from hidori_core.compat.typing import Literal
from hidori_core.modules.base import Module
from hidori_core.schema.base import Schema
from hidori_core.schema.constraints import Requires
from hidori_core.utils import Messenger

DNF_STATE_INSTALLED = "installed"
DNF_STATE_UPGRADED = "upgraded"


def condition_require_package_if_state_installed(data: Dict[str, Any]) -> bool:
    return data.get("state") == DNF_STATE_INSTALLED


class DnfSchema(Schema):
    state: Literal["upgraded", "installed"] = Requires(
        ["package"], data_conditions=[condition_require_package_if_state_installed]
    )
    package: Optional[str]


class DnfModule(Module, name="dnf", schema_cls=DnfSchema):
    def execute(
        self, validated_data: Dict[str, Any], messenger: Messenger
    ) -> Dict[str, str]:
        base = dnf.Base()
        base.read_all_repos()
        base.fill_sack()
        package_name = validated_data.get("package")

        if validated_data["state"] == "installed":
            # TODO: Only allow package to be a list and name it packages
            # TODO: Verify that package exists in repos
            # TODO: Allow to provide version with `package:version` or
            # `package:latest` notation
            installed_query = base.sack.query().installed()
            if installed_query.filter(name=package_name):
                messenger.queue_success(f"package {package_name} is already installed")
                return {"state": "unaffected"}

            base.install(package_name)
            base.resolve()
            base.download_packages(base.transaction.install_set)

            for package in base.transaction.install_set:
                # TODO: verbose only, messenger should accept boolean parameter verbose
                messenger.queue_info(
                    f"will install {package.name} to version "
                    f"{package.version}-{package.release}"
                )

            base.do_transaction()
            messenger.queue_affected(f"package {package_name} has been installed")
            return {"state": "affected"}

        if validated_data["state"] == "upgraded" and package_name is None:
            base.upgrade_all()
            base.resolve()
            base.download_packages(base.transaction.install_set)

            if not base.transaction.install_set:
                messenger.queue_success("all packages are up to date")
                return {"state": "unaffected"}

            package_count = len(base.transaction.install_set)
            for package in base.transaction.install_set:
                # TODO: verbose only, messenger should accept boolean parameter verbose
                messenger.queue_info(
                    f"will upgrade {package.name} to version "
                    f"{package.version}-{package.release}"
                )

            base.do_transaction()
            messenger.queue_affected(f"performed upgrade of {package_count} packages")
            return {"state": "affected"}

        messenger.queue_error("internal error")
        return {"state": "error"}



# Kinda necessary because of Debian 10 which should be supported until June 2024,
# but since Python 3.7 will EOL around July 2023 hidori will drop Debian 10
# at that time as well.
# TODO: Remove in July 2023 when python 3.7 will probably EOL

# 3.8+:
try:
    from typing import Literal
# 3.7:
except ImportError:
    from typing_extensions import Literal  # type: ignore


__all__ = ["Literal"]



from __future__ import annotations

import argparse
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

TD = TypeVar("TD", bound="BaseData")

COMMAND_REGISTRY: defaultdict[str, dict[str, type[Command[Any]]]] = defaultdict(dict)


@dataclass
class BaseData:
    subparser_name: str


class Command(Generic[TD]):
    name: str
    data_cls: type[TD]
    backend_parse_content: bool = False

    def __init_subclass__(cls) -> None:
        app_name, cmd_name, _ = re.findall(".[^A-Z]*", cls.__name__)
        app_name = f"hidori-{app_name.lower()}"
        cmd_name = cmd_name.lower()

        if cmd_name in COMMAND_REGISTRY:
            raise RuntimeError()
        cls.name = cmd_name
        COMMAND_REGISTRY[app_name][cmd_name] = cls

    def __init__(self, subparser: argparse.ArgumentParser) -> None:
        for field_name, field in self.data_cls.__dataclass_fields__.items():
            if field_name == "subparser_name":
                continue

            kwargs = {}
            help_text = field.metadata.get("help", "")
            is_positional = field.metadata.get("is_positional", True)
            if help_text:
                kwargs["help"] = help_text

            if field.type in (bool, "bool"):
                subparser.add_argument(f"--{field.name}", action="store_true", **kwargs)

            if field.type in (str, "str"):
                name = field.name if is_positional else f"--{field_name}"
                subparser.add_argument(name, type=str, **kwargs)

    def execute(self, data: TD) -> None:
        ...
